from .node import Node


class Task(Node):
    def __init__(self, title, path, docname, dependencies, userdata,
                 responsible, initial_estimate, spent, percent_done):
        super().__init__(
            title=title, 
            path=path, 
            docname=docname, 
            dependencies=dependencies, 
            userdata=userdata)
        self.initial_estimate = float(initial_estimate)
        self.spent = float(spent)
        self.percent_done = float(percent_done)
        self.responsible = responsible

    def __str__(self):
        return 'Task:'+super().__str__()



from ..soup import Soup
from ..topic import Topic
from ..exercise import Exercise
from ..task import Task
from ..group import Group
from .. import errors

from sphinx.util import logging
_logger = logging.getLogger(__name__)


def _prepare_app(app):
    if hasattr(app, 'ot_soup'):
        raise OpenTrainingError('Soup already created, cannot add one more element')
    if not hasattr(app.env, 'ot_elements'):
        app.env.ot_elements = {}

def sphinx_add_element(app, element):
    _prepare_app(app)
    app.env.ot_elements[element.docname] = element

def sphinx_purge_doc(app, env, docname):
    if hasattr(env, 'ot_elements'):
        env.ot_elements.pop(docname, None)

def sphinx_create_soup(app):
    if hasattr(app, 'ot_soup'):
        return

    app.ot_soup = Soup()
    for element in app.env.ot_elements.values():
        app.ot_soup.add_element(element)

    try:
        app.ot_soup.commit()
    except errors.CompoundError as e:
        for err in e.errors:
            if isinstance(err, errors.ElementError):
                _logger.warning(str(err), location=err.element.userdata)
            else:
                assert False, f'hm. cannot handle {err}'
            



from . import errors


class Element:
    def __init__(self, title, path, docname, userdata):
        _verify_is_path(path)

        self.title = title
        self.docname = docname
        self.userdata = userdata
        if path:
            self._requested_path = path
        else:   # root group; no parent
            self.parent = None

    def __str__(self):
        if hasattr(self, '_requested_path'):
            return f'{self._requested_path} (uncommitted)'
        else:
            return f'{self.path}'

    @property
    def path(self):
        if hasattr(self, '_requested_path'):
            raise errors.OpenTrainingError(f'{self} is not yet committed (path not known)')
        if self.parent:
            return self.parent.path + [self.parent.element_name(self)]
        else:
            return []


def _verify_is_path(path):
    if type(path) not in (list, tuple):
        raise errors.BadPath(f'Not a valid path: {path} is neither list nor tuple')
        
    for elem in path:
        if type(elem) is not str:
            raise errors.BadPath(f'Not a valid path: {path} ({elem} is not str)')
        if not elem.isidentifier():
            raise errors.BadPath(f'Not a valid path: {path} ({elem} is not an identifier)')
    



from .sphinxglue import topic
from .sphinxglue import exercise
from .sphinxglue import task
from .sphinxglue import group
from .sphinxglue import graph
from .sphinxglue import soup
from .sphinxglue import dia

# from sphinx.util.docutils import SphinxDirective
# from sphinx.util.nodes import set_source_info
# from sphinx.util import logging
# from docutils import nodes


# logger = logging.getLogger(__name__)

def setup(app):
    app.connect('env-purge-doc', soup.sphinx_purge_doc)

    topic.setup(app)
    exercise.setup(app)
    task.setup(app)
    group.setup(app)
    graph.setup(app)
    dia.setup(app)





from . import utils
from . import soup
from .graph import TopicGraphNode
from .. import errors
from ..task import Task

from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import set_source_info
from sphinx.util import logging
from docutils import nodes

logger = logging.getLogger(__name__)


def setup(app):
    app.add_directive('ot-task', _TaskDirective)
    app.connect('doctree-read', _ev_doctree_read__extract_tasknodes)

def _ev_doctree_read__extract_tasknodes(app, doctree):
    try:
        docname = app.env.docname
        task_nodes = list(doctree.traverse(_TaskNode))
        if len(task_nodes) > 1:
            raise errors.OpenTrainingError(f'{docname} contains multiple tasks')

        for n in task_nodes:
            soup.sphinx_add_element(app, Task(
                docname=docname,
                title=utils.get_document_title(docname, doctree),
                path=n.path, 
                userdata=n,
                dependencies=n.dependencies,
                responsible=n.responsible,
                initial_estimate=n.initial_estimate,
                spent=n.spent,
                percent_done=n.percent_done,
            ))
            n.replace_self([])
    except Exception:
        logger.exception(f'{docname}: cannot extract task nodes')
        raise
        
class _TaskNode(nodes.Element):
    def __init__(self, path, dependencies,
                 responsible, initial_estimate, spent, percent_done):
        super().__init__(self)
        self.title = None
        self.path = path
        self.dependencies = dependencies
        self.responsible = responsible
        self.initial_estimate = initial_estimate
        self.spent = spent
        self.percent_done = percent_done

class _TaskDirective(SphinxDirective):
    required_arguments = 1   # path

    option_spec = {
        'dependencies': utils.list_of_element_path,
        'responsible': str,
        'initial-estimate': int,
        'spent': int,
        'percent-done': int,
    }

    def run(self):
        path = utils.element_path(self.arguments[0].strip())
        dependencies = self.options.get('dependencies', [])
        responsible = self.options.get('responsible', '')
        initial_estimate = self.options.get('initial-estimate', 0)
        spent = self.options.get('spent', 0)
        percent_done = self.options.get('percent-done', 0)

        task = _TaskNode(path=path, 
                         dependencies=dependencies,
                         responsible=responsible,
                         initial_estimate=initial_estimate, 
                         spent=spent, 
                         percent_done=percent_done)
        task.document = self.state.document
        set_source_info(self, task)

        return [task]



from docutils import nodes


def element_path(pathstr):
    return [c.strip() for c in pathstr.split('.')]

def list_of_element_path(pathliststr):
    paths = [p.strip() for p in pathliststr.split(',')]
    return [element_path(p) for p in  paths]

def get_document_title(docname, doctree):
    for section in doctree.traverse(nodes.section):
        break
    else:
        raise Exception(f'{docname}: no <section> found')
    for title in section.traverse(nodes.title):
        return title.astext()
    else:
        raise Exception(f'{docname}: first <section> has no <title>')

